"""Import subclasses here to init registry in base class."""

from genai_bench.sampling.base import Sampler
from genai_bench.sampling.image import ImageSampler
from genai_bench.sampling.text import TextSampler
